# ANALYZE FONT NAME
package Autofont;
use Autofont::Encoding;

$VERSION = '0.96';

$suffix_separator = '--';

sub complain ($) {
    print STDERR "mktextfm: warning: ", @_, "\n";
}

sub new ($$;$) {
    my($class, $fontname, $fontmap) = @_;
    my($font, $suffix);

    # create self
    $font = { };
    bless $font, $class;
    $font->{'basename'} = $fontname;

    # take off explicit appliques
    ($fontname, $suffix) = ($fontname =~ /\A(.*?)(--.*|)\Z/);

    # handle encodings embedded in the name
    if ($fontname =~ /^r(p.*)$/) {
	$fontname = $1;
	$font->{'want_raw'} = 1;
    }
    if ($fontname =~ /^(.*)8r$/) {
	($fontname, $suffix) = ($1, "${suffix}--8r");
    }

    # access %fontmap
    if (defined($fontmap) && exists($fontmap->{$fontname})) {
	$fontname = $fontmap->{$fontname};
	($fontname, $suffix) = ($1, "${suffix}$2")
	    if $fontname =~ /\A(.*?)(--.*)\Z/;
    }

    # separate appliques
    my(@appliques) = split(/--/, $suffix);

    # multiple master arguments
    $font->{'mmname'} = $fontname;
    $font->{'mm'} = 0;
    if ($fontname =~ /^([^_]*)_([0-9._]+)$/) {
	$fontname = $1;
	$font->{'mm'} = 1;
	$font->{'mmv'} = [split /_/, $2];
    }

    # set name, analyze appliques, and return
    $font->{'name'} = $fontname;
    $font->_analyze_appliques(@appliques);
    
    $font;
}

sub _analyze_appliques ($@) {
    my($font, @a) = @_;
    my($a, $encoding);
  
    # analyze appliques
    foreach $a (@a) {

	if ($a eq '') {
	    # blank appliques OK
    
	} elsif ($a =~ /^sl([-+\d.]+)$/) {
	    complain("more than one slant") if exists $font->{'slant'};
	    $font->{'slant'} = $1 / 1000;
      
	} elsif ($a =~ /^ex([\d.]+)$/ || $a =~ /^cn([\d.]+)$/) {
	    complain("more than one expansion") if exists $font->{'expand'};
	    $font->{'expand'} = $1 / 100;

	} elsif ($a =~ /^fsc$/) {
	    $font->{'fake_sc'} = 1;

	} elsif ($a =~ /^osf$/) {
	    complain("figure type already specified") if exists($font->{'tif'}) || exists $font->{'rawexpf'};
	    $font->{'osf'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^tif$/) {
	    complain("figure type already specified") if exists($font->{'osf'}) || exists $font->{'rawexpf'};
	    $font->{'tif'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^rawexpf$/) {
	    complain("figure type already specified") if exists($font->{'osf'}) || exists $font->{'tif'};
	    $font->{'rawexpf'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^sc$/) {
	    $font->{'sc'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^swc$/) {
	    $font->{'swash_caps'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^f$/) {
	    $font->{'f_ligs'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^nof$/) {
	    $font->{'no_f_ligs'} = $font->{'built'} = 1;
	    
	} elsif ($a =~ /^lsp([-+\d.]+)$/) {
	    $font->{'letterspace'} = $1;
	    $font->{'built'} = 1;

	} elsif ($a =~ /^spstr([-+\d.]+)$/) {
	    $font->{'spacestretch'} = $1 / 100;
	    $font->{'built'} = 1;
	    
	} elsif ($a =~ /^0$/) {
	    $font->{'want_raw'} = 1;
      
	} elsif ($a =~ /^E(.*)$/) {
	    complain("more than one encoding") if exists $font->{'encoding'};
	    $encoding = $1;
	    
	} else {
	    complain("more than one encoding") if exists $font->{'encoding'};
	    $encoding = $a;
	}
    
    }

    complain("built font cannot be raw") if exists $font->{'built'} && exists $font->{'want_raw'};
    
    # find encoding
    $encoding = $::ENV{"TEXDEFAULTENCODING"} if !defined $encoding;
    $encoding = '8r' if !defined $encoding;
    $font->{'encoding'} = Autofont::Encoding->named($encoding);
    die "cannot find $encoding encoding" if !defined $font->{'encoding'};

    # set suffix
    $font->{'suffix'} = "--E" . $encoding;
    $font->{'suffix'} .= "--sl" . $font->{'slant'}*1000
	if exists $font->{'slant'};
    $font->{'suffix'} .= "--ex" . $font->{'expand'}*100
	if exists $font->{'expand'};
    $font->{'suffix'} .= "--fsc"
	if exists $font->{'fake_sc'};

    $font;
}


sub is_autofont ($) {
    1;
}

sub mm ($) {
    my($font) = @_;
    $font->{'mm'};
}

sub encoding ($) {
    my($font) = @_;
    $font->{'encoding'};
}

sub suffixed_name ($) {
    my($font) = @_;
    return $font->{'name'} . $font->{'suffix'};
}

sub replace_base_name ($$) {
    my($font, $basename) = @_;

    # might need to "macintosh" the font name
    $basename = _mac_font_name_abbrev($basename)
	if $font->{'name_transform'} eq 'mac';
    
    if ($font->{'mm'}) {
	$basename . '_' . join('_', @{$font->{'mmv'}}) . '_' . $font->{'suffix'};
    } else {
	$basename . $font->{'suffix'};
    }
}

sub _find_mac_key ($$) {
    my($db, $name) = @_;
    my($k, $v);
    keys %$db;			# clear iterator, fucking Perl
    
    my($head) = ($name =~ /^([A-Za-z][a-z]*)/);
    my($head_len) = length($head);
    return undef if $head_len == 0 || $head_len > 5;
    
    while (($k, $v) = each %$db) {
	next if substr($k, 0, $head_len) ne $head;
	next if _mac_font_name_abbrev($k) ne $name;
	return $k;
    }
    
    return undef;
}

sub psname ($;$) {
    my($font, $psres) = @_;

    return $font->{'psname'} if exists($font->{'psname'});

    my($name, $psname) = $font->{'name'};
    my($metrics_db) = ($font->mm ? $psres->{'FontAMFM'} : $psres->{'FontAFM'});
    my($outline_db) = $psres->{'FontOutline'};

    # check the font name and "un-Macintoshed" versions of the font name
    $font->{'name_transform'} = 'unknown';
    if (defined($outline_db->{$name}) || defined($metrics_db->{$name})) {
	$psname = $name;
    } else {
	$psname = _find_mac_key($metrics_db, $name);
	$psname = _find_mac_key($outline_db, $name) if !defined $psname;
	$font->{'name_transform'} = 'mac' if defined $psname;
    }
    
    if (defined($psname)) {
	$font->{'psname'} = $font->{'mmpsname'} = $psname;
	$font->{'mmpsname'} .= '_' . join('_', @{$font->{'mmv'}}) . '_'
	    if $font->mm;
	$font->{'metrics_file'} = $metrics_db->{$psname};
	$font->{'outline_file'} = $psres->{'FontOutline'}->{$psname};
    }
    
    $psname;
}

sub familyname ($$) {
    my($font, $psres) = @_;
    $psres->fontfamily($font->psname($psres));
}

sub ps_expert_fullname ($$) {
    my($font, $psres) = @_;
    $font->is_autofont || die;
    return undef if !$font->psname($psres);
    my($metrics_db) = ($font->mm ? $psres->{'FontAMFM'} : $psres->{'FontAFM'});
    my($psfn, $epfn) = $font->psname;
    while (1) {
	$epfn = $psfn; $epfn =~ s/-/Exp-/;
	last if $epfn ne $psfn && exists $metrics_db->{$epfn};
	$epfn = $psfn; $epfn =~ s/-/Expert-/;
	last if $epfn ne $psfn && exists $metrics_db->{$epfn};
	$epfn = $psfn; $epfn =~ s/(MM-.*)/$1Ep/; $epfn =~ s/MM$/MM-Ep/;
	last if $epfn ne $psfn && exists $metrics_db->{$epfn};
	$epfn = $psfn; $epfn =~ s/([A-Z][A-Z])($|-)/Expert$1$2/;
	last if $epfn ne $psfn && exists $metrics_db->{$epfn};
	$epfn = $psfn . "Expert";
	last if $epfn ne $psfn && exists $metrics_db->{$epfn};
	if ($psfn =~ s/Alt(ernative)?//) {
	    # loop again without the Alt
	} elsif ($psfn =~ s/Semibold/SemiBold/) {
	    # some fonts (Bembo) have this weird problem too: expert is called
	    # SemiBold, although normal is called Semibold.
	} else {
	    $epfn = $font->suffixed_name();
	    $epfn =~ s{--E(.*?)($|--)}{--Easexp$2};
	    return $epfn;
	}
    }
    $epfn = $font->replace_base_name($epfn);

    # change encoding
    if (!($epfn =~ s{--Eeka($|--)}{--Eekx$1})) {
	$epfn =~ s{--E(.*?)($|--)}{--Easexp$2};
    }
    
    $epfn;
}

sub ps_osf_fullname ($$) {
    my($font, $psres) = @_;
    $font->is_autofont || die;
    return undef if !$font->psname($psres);
    my($metrics_db) = ($font->mm ? $psres->{'FontAMFM'} : $psres->{'FontAFM'});
    my($psfn, $osfn) = $font->psname;
    {
	$osfn = $psfn . "SC";
	last if exists $metrics_db->{$osfn};
	$osfn = $psfn . "OsF";
	last if exists $metrics_db->{$osfn};
	$osfn = $psfn . "-SC";
	last if exists $metrics_db->{$osfn};
	$osfn = $psfn . "-OsF";
	last if exists $metrics_db->{$osfn};
	return $font->ps_expert_fullname($psres);
    };
    $font->replace_base_name($osfn);
}

sub ps_swash_fullname ($$) {
    my($font, $psres) = @_;
    $font->is_autofont || die;
    return undef if !$font->psname($psres);
    my($metrics_db) = ($font->mm ? $psres->{'FontAMFM'} : $psres->{'FontAFM'});
    my($psfn, $swfn) = $font->psname;
    while (1) {
	$swfn = $psfn; $swfn =~ s/-/-SwashCaps/;
	last if $swfn ne $psfn && exists $metrics_db->{$swfn};
	$swfn = $psfn; $swfn =~ s/-/-Swash/;
	last if $swfn ne $psfn && exists $metrics_db->{$swfn};
	$swfn = $psfn . "-Swash";
	last if $swfn ne $psfn && exists $metrics_db->{$swfn};
	$swfn = $psfn; $swfn =~ s/(MM-.*)/$1Sw/; $swfn =~ s/MM$/MM-Sw/;
	last if $swfn ne $psfn && exists $metrics_db->{$swfn};
	return undef if !($psfn =~ s/Alt(ernative)?//)
	    && !($psfn =~ s/-?It(alic)?//);
    };
    $font->replace_base_name($swfn);
}

sub ps_fullname_exists ($$$) {
    my($font, $fontname, $psres) = @_;
    $font->is_autofont || die;
    return undef if !$font->psname($psres);
    my($metrics_db) = ($font->mm ? $psres->{'FontAMFM'} : $psres->{'FontAFM'});
    exists $metrics_db->{$fontname};
}

sub psfonts_map_line ($;$) {
    my($font, $psres) = @_;
    die if !$font->psname($psres);
    my($t) = $font->{'mmpsname'} . " ";
    my($mod) = '';

    # add encoding
    my($enc) = $font->{'encoding'};
    if ($enc && $enc->psname) {
	$mod .= " " . $enc->psname . " ReEncodeFont";
    }

    # add slant, etc.
    $mod .= " $font->{'slant'} SlantFont" if $font->{'slant'};
    $mod .= " $font->{'extend'} ExtendFont" if $font->{'extend'};
    
    $t .= '"' . substr($mod, 1) . '"' if $mod ne '';
    $t .= " <[" . $enc->filename if $enc && $enc->filename;
    $t .= ($font->{'mm'} ? " <<" : " <") . $font->{'outline_file'}
	if (defined($font->{'outline_file'})
	    && !exists $::no_download_outline{$font->{'psname'}});

    $t =~ s|/+|/|g;

    $t;
}

sub _mac_font_name_abbrev ($) {
    my($fn) = @_;

    my($start, $end) = ($fn =~ /^([A-Z][a-z]*)(.*)/);
    $start = "", $end = $fn if !defined $start;
    $start =~ s/([A-Z][a-z]{4}).*/$1/;
    $end =~ s/([A-Za-z][a-z]{2})[a-z]*/$1/g;
    $end =~ tr/-//d;

    $start . $end;
}

1;
